/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.List;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.server.RepresentationModelAssembler;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.downloadtoken.DownloadTokenController;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserManagerUrlProvider;
import de.ozgcloud.alfa.common.user.UserRole;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
public class RootModelAssembler implements RepresentationModelAssembler<Root, EntityModel<Root>> {

	static final String REL_DOWNLOAD_TOKEN = "downloadToken";
	static final String REL_CURRENT_USER = "currentUser";
	static final String REL_DOCUMENTATIONS = "documentations";

	private final RootViewLinkHandler viewLinkHandler;
	private final CurrentUserService currentUserService;
	private final UserManagerUrlProvider userManagerUrlProvider;

	@Value("${ozgcloud.user-assistance.documentation.url:#{null}}")
	private String documentationUrl;

	@Override
	public EntityModel<Root> toModel(Root root) {
		var model = EntityModel.of(root)
				.addAllIf(hasAnyRole(), () -> List.of(
						linkTo(RootController.class).withSelfRel(),
						linkTo(DownloadTokenController.class).withRel(REL_DOWNLOAD_TOKEN)))
				.addAllIf(userManagerUrlProvider.isConfiguredForUserProfile(), this::buildCurrentUserLink)
				.add(buildUserAssistanceDocumentationUrl());

		if (hasAnyRole()) {
			viewLinkHandler.addViewLinks(model, currentUserService.findUserId());
		}

		return model;
	}

	boolean hasAnyRole() {
		return hasVerwaltungRole() || currentUserService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER);
	}

	boolean hasVerwaltungRole() {
		return currentUserService.hasRole(UserRole.VERWALTUNG_USER) || currentUserService.hasRole(UserRole.VERWALTUNG_POSTSTELLE);
	}

	private List<Link> buildCurrentUserLink() {
		return currentUserService.findUserId().map(userId -> Link.of(String.format(userManagerUrlProvider.getUserProfileTemplate(), userId))
				.withRel(REL_CURRENT_USER)).stream().toList();
	}

	private List<Link> buildUserAssistanceDocumentationUrl() {
		return Stream.of(documentationUrl).filter(StringUtils::isNotBlank).map(href -> Link.of(href, REL_DOCUMENTATIONS)).toList();
	}
}