/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.forwarding;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;

import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;
import org.springframework.hateoas.server.RepresentationModelProcessor;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.FeatureToggleProperties;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
class ForwardingVorgangWithEingangProcessor implements RepresentationModelProcessor<EntityModel<VorgangWithEingang>> {

	static final LinkRelation REL_FORWARD_BY_OZGCLOUD = LinkRelation.of("forwardByOzgCloud");
	static final LinkRelation REL_FORWARD_BY_EMAIL = LinkRelation.of("forwardByEmail");

	private final FeatureToggleProperties featureToggleProperties;
	private final CurrentUserService userService;

	@Override
	public EntityModel<VorgangWithEingang> process(EntityModel<VorgangWithEingang> model) {

		Optional.ofNullable(model.getContent())
				.ifPresent(vorgang -> model
						.addIf(isForwardableByOzgCloud(vorgang), () -> buildForwardByOzgCloudLink(vorgang))
						.addIf(isForwardableByEmail(), () -> buildForwardByEmailLink(vorgang)));
		return model;
	}

	boolean isForwardableByOzgCloud(VorgangWithEingang vorgang) {
		return featureToggleProperties.isForwardByOzgCloudEnabled() && isStatusNeu(vorgang);
	}

	private boolean isStatusNeu(VorgangWithEingang vorgang) {
		return vorgang.getStatus() == VorgangStatus.NEU;
	}

	private Link buildForwardByOzgCloudLink(VorgangWithEingang vorgang) {
		return linkTo(methodOn(ForwardByVorgangCommandController.class).createCommand(vorgang.getId(),
				vorgang.getVersion(), null)).withRel(REL_FORWARD_BY_OZGCLOUD);
	}

	boolean isForwardableByEmail() {
		return userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER);
	}

	private Link buildForwardByEmailLink(VorgangWithEingang vorgang) {
		return linkTo(methodOn(ForwardingController.class).findByVorgangId(vorgang.getId())).withRel(REL_FORWARD_BY_EMAIL);
	}

}
