/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.loeschanforderung;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;

import org.springframework.hateoas.EntityModel;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItemService;
import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandController;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandStatus;
import de.ozgcloud.alfa.common.command.CreateCommand;
import de.ozgcloud.alfa.common.command.StatusPatch;
import de.ozgcloud.alfa.common.errorhandling.OrderNotAllowedException;
import de.ozgcloud.alfa.vorgang.VorgangController;
import lombok.RequiredArgsConstructor;

@RestController
@RequestMapping(LoeschAnforderungCommandController.BASE_PATH)
@RequiredArgsConstructor
public class LoeschAnforderungCommandController {

	static final String BASE_PATH = "/api/loeschanforderungs/{loeschAnforderungId}/commands"; // NOSONAR

	private final LoeschAnforderungService loeschAnforderungService;

	private final VorgangAttachedItemService vorgangAttachedItemService;

	private final VorgangController vorgangController;

	private final CommandController commandController;

	@PreAuthorize("@loeschAnforderungPreAuthorizeHandler.hasPermissionForOrder(#loeschAnforderungId, #command.order)")
	@PostMapping
	public ResponseEntity<Void> createCommand(@PathVariable String loeschAnforderungId, @RequestBody CreateCommand command) {
		var createdCommand = executeCommand(command, loeschAnforderungId);

		return ResponseEntity.created(linkTo(CommandController.class).slash(createdCommand.getId()).toUri()).build();
	}

	private Command executeCommand(CreateCommand command, String loeschAnforderungId) {
		return switch (command.getCommandOrder()) {
			case VORGANG_LOESCHEN -> loeschAnforderungService.vorgangLoeschen(command, loeschAnforderungId);
			case LOESCH_ANFORDERUNG_ZURUECKNEHMEN -> loeschAnforderungZuruecknehmen(loeschAnforderungId);
			default -> throw new OrderNotAllowedException(command.getOrder());
		};
	}

	private Command loeschAnforderungZuruecknehmen(String loeschAnforderungId) {
		var vorgangAttachedItem = vorgangAttachedItemService.getById(loeschAnforderungId);
		var vorgang = vorgangController.getVorgang(vorgangAttachedItem.getVorgangId());
		return loeschAnforderungService.zuruecknehmen(vorgangAttachedItem, vorgang);
	}

	@PatchMapping("{commandId}")
	public ResponseEntity<EntityModel<Command>> revoke(@PathVariable String loeschAnforderungId, @PathVariable String commandId,
			@RequestBody StatusPatch patch) {
		if (patch.getStatus() == CommandStatus.REVOKED) {
			var command = commandController.getById(commandId).getContent();

			if (Optional.ofNullable(command).map(Command::getCommandOrder).orElse(null) == CommandOrder.VORGANG_ZUM_LOESCHEN_MARKIEREN) {
				return handleVorgangZumLoeschenMarkierenRevokeCommand(loeschAnforderungId, command, patch);
			}
			if (Optional.ofNullable(command).map(Command::getCommandOrder).orElse(null) == CommandOrder.LOESCH_ANFORDERUNG_ZURUECKNEHMEN) {
				return handleLoeschAnforderungZuruecknehmenRevokeCommand(command, patch);
			}
		}
		return ResponseEntity.status(HttpStatus.FORBIDDEN).build();
	}

	ResponseEntity<EntityModel<Command>> handleVorgangZumLoeschenMarkierenRevokeCommand(String loeschAnforderungId, Command command,
			StatusPatch patch) {
		loeschAnforderungService.deleteLoeschAnforderung(loeschAnforderungId);

		return commandController.revoke(command.getId(), patch);
	}

	ResponseEntity<EntityModel<Command>> handleLoeschAnforderungZuruecknehmenRevokeCommand(Command command, StatusPatch patch) {
		loeschAnforderungService.reCreateLoeschAnforderung(command);

		var changeStatusCommandId = loeschAnforderungService.getChangeStatusCommandId(command);
		return commandController.revoke(changeStatusCommandId, patch);
	}
}