/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.wiedervorlage;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.function.Predicate;
import java.util.stream.Stream;

import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.server.RepresentationModelAssembler;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.binaryfile.BinaryFileController;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageCommandController.WiedervorlageCommandByVorgangController;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageController.AttachmentsByWiedervorlageController;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
class WiedervorlageModelAssembler implements RepresentationModelAssembler<Wiedervorlage, EntityModel<Wiedervorlage>> {

	static final String REL_EDIT = "edit";
	static final String REL_ERLEDIGEN = "erledigen";
	static final String REL_WIEDEREROEFFNEN = "wiedereroeffnen";
	static final String REL_ATTACHMENTS = "attachments";

	static final String REL_CREATE = "create-wiedervorlage";
	static final String REL_UPLOAD_FILE = "uploadFile";

	private static final Predicate<Wiedervorlage> ALLOW_ERLEDIGEN = wiedervorlage -> !wiedervorlage.isDone();
	private static final Predicate<Wiedervorlage> ALLOW_WIEDEREROEFFNEN = Wiedervorlage::isDone;
	private static final Predicate<Wiedervorlage> HAS_ATTACHMENTS = wiedervorlage -> !wiedervorlage.getAttachments().isEmpty();
	public static final String WIEDERVORLAGE_ATTACHMENT_PATH = "wiedervorlageAttachment";
	public static final String FILE_PATH = "file";

	private final VorgangController vorgangController;

	@Override
	public EntityModel<Wiedervorlage> toModel(Wiedervorlage wiedervorlage) {
		var vorgang = vorgangController.getVorgang(wiedervorlage.getVorgangId());
		return buildModel(wiedervorlage, vorgang);
	}

	EntityModel<Wiedervorlage> buildModel(Wiedervorlage wiedervorlage, VorgangWithEingang vorgang) {
		var selfLink = linkTo(WiedervorlageController.class).slash(wiedervorlage.getId());
		var commandLink = linkTo(
				methodOn(WiedervorlageCommandController.class).updateWiedervorlage(null, wiedervorlage.getId(), wiedervorlage.getVersion()));

		return EntityModel.of(wiedervorlage)
				.add(selfLink.withSelfRel())
				.addIf(vorgangController.isEditable(vorgang), () -> commandLink.withRel(REL_EDIT))
				.addIf(ALLOW_ERLEDIGEN.test(wiedervorlage), () -> commandLink.withRel(REL_ERLEDIGEN))
				.addIf(ALLOW_WIEDEREROEFFNEN.test(wiedervorlage), () -> commandLink.withRel(REL_WIEDEREROEFFNEN))
				.addIf(HAS_ATTACHMENTS.test(wiedervorlage),
						() -> linkTo(methodOn(AttachmentsByWiedervorlageController.class).getAttachments(wiedervorlage.getId()))
								.withRel(REL_ATTACHMENTS));
	}

	public CollectionModel<EntityModel<Wiedervorlage>> toCollectionModel(Stream<Wiedervorlage> entities, String vorgangId) {
		var vorgang = vorgangController.getVorgang(vorgangId);
		return CollectionModel.of(entities.map(wiedervorlage -> buildModel(wiedervorlage, vorgang)).toList())
				.add(linkTo(WiedervorlageController.class).withSelfRel())
				.add(linkTo(BinaryFileController.class).slash(vorgangId).slash(WIEDERVORLAGE_ATTACHMENT_PATH).slash(FILE_PATH)
						.withRel(REL_UPLOAD_FILE))
				.addIf(vorgangController.isEditable(vorgang),
						() -> linkTo(methodOn(WiedervorlageCommandByVorgangController.class).createWiedervorlage(null, vorgangId))
								.withRel(REL_CREATE));
	}
}