/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.bescheid;

import static de.ozgcloud.alfa.bescheid.BescheidModelAssembler.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Collections;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.web.util.UriTemplate;

import de.ozgcloud.alfa.bescheid.BescheidController.BescheidByVorgangController;
import de.ozgcloud.alfa.common.binaryfile.FileId;
import de.ozgcloud.alfa.common.command.CommandController;
import de.ozgcloud.alfa.postfach.PostfachMailController;
import de.ozgcloud.alfa.vorgang.Eingang;
import de.ozgcloud.alfa.vorgang.EingangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.EingangTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHeadTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangProperties;
import de.ozgcloud.alfa.vorgang.VorgangPropertyTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class BescheidModelAssemblerTest {

	@Spy
	@InjectMocks
	private BescheidModelAssembler assembler;

	@Mock
	private BescheidService bescheidService;

	@Mock
	private VorgangProperties vorgangProperties;

	@Mock
	private VorgangController vorgangController;
	@Mock
	private PostfachMailController postfachMailController;

	@Nested
	class TestToModel {

		private final Bescheid bescheid = BescheidTestFactory.create();
		private final VorgangWithEingang vorgangWithEingang = VorgangWithEingangTestFactory.create();

		@BeforeEach
		void setUp() {
			when(vorgangController.getVorgang(VorgangHeaderTestFactory.ID)).thenReturn(vorgangWithEingang);
		}

		@Test
		void shouldHaveSelfLink() {
			var model = callToModel();

			assertThat(model.getLink(IanaLinkRelations.SELF)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(BescheidController.PATH + "?vorgangId=" + BescheidTestFactory.VORGANG_ID);
		}

		@Test
		void shouldHaveDeleteLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_DELETE)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(new UriTemplate(CommandController.CommandByRelationController.COMMAND_BY_RELATION_PATH)
							.expand(VorgangHeaderTestFactory.ID, BescheidTestFactory.ID, BescheidTestFactory.VERSION).toString());
		}

		@Test
		void shouldHaveUploadBescheidFileLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_UPLOAD_BESCHEID_FILE)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(uploadFileLink("bescheidFile"));
		}

		@Test
		void shouldHaveUploadAttachmentLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_UPLOAD_ATTACHMENT)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(uploadFileLink("bescheidAttachment"));
		}

		@ParameterizedTest
		@NullAndEmptySource
		void shouldNotHaveAttachmentsLinkIfNoAttachmentsAvailable(List<FileId> attachments) {
			var bescheid = BescheidTestFactory.createBuilder().attachments(attachments).build();

			var model = callToModel(bescheid);

			assertThat(model.getLink(REL_ATTACHMENTS)).isEmpty();
		}

		@Test
		void shouldHaveAttachmentsLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_ATTACHMENTS)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(String.format("%s/%s/attachments", BescheidController.PATH, BescheidTestFactory.ID));
		}

		@Test
		void shouldHaveUpdateLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_UPDATE)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(createCommandLink());
		}

		@Test
		void shouldHaveCreateDocumentLink() {
			doReturn(true).when(assembler).canCreateBescheidDocumentAutomatically(vorgangWithEingang);

			var model = callToModel();

			assertThat(model.getLink(REL_CREATE_DOCUMENT)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(createCommandLink());
		}

		@Test
		void shoulNotdHaveCreateDocumentLink() {
			doReturn(false).when(assembler).canCreateBescheidDocumentAutomatically(vorgangWithEingang);

			var model = callToModel();

			assertThat(model.getLink(REL_CREATE_DOCUMENT)).isEmpty();
		}

		@Test
		void shouldHaveCreateDocumentFromFileLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_CREATE_DOCUMENT_FROM_FILE)).isPresent().get().extracting(Link::getHref)
					.isEqualTo(createCommandLink());
		}

		@Test
		void shouldHaveBescheidenUndSendenLink() {
			when(postfachMailController.isPostfachConfigured()).thenReturn(true);

			var model = callToModel();

			assertThat(model.getLink(REL_BESCHEIDEN_UND_SENDEN))
					.isPresent().get()
					.extracting(Link::getHref).isEqualTo(bescheidenUndSendenLink());
		}

		@Test
		void shouldNotHaveBescheidenUndSendenLinkOnNotConfiguredPostfach() {
			when(postfachMailController.isPostfachConfigured()).thenReturn(false);

			var model = callToModel();

			assertThat(model.getLink(REL_BESCHEIDEN_UND_SENDEN)).isEmpty();
		}

		@Test
		void shouldNotHaveBescheidenUndSendenLinkOnVorgangWithoutServiceKonto() {
			var vorgang = VorgangWithEingangTestFactory.createBuilder()
					.header(VorgangHeadTestFactory.createBuilder().serviceKonto(null).build())
					.build();
			when(vorgangController.getVorgang(VorgangHeaderTestFactory.ID)).thenReturn(vorgang);
			when(postfachMailController.isPostfachConfigured()).thenReturn(true);

			var model = callToModel();

			assertThat(model.getLink(REL_BESCHEIDEN_UND_SENDEN)).isEmpty();
		}

		@Test
		void shouldHaveBescheidenLink() {
			var model = callToModel();

			assertThat(model.getLink(REL_BESCHEIDEN))
					.isPresent().get()
					.extracting(Link::getHref).isEqualTo(createCommandLink());
		}

		@Test
		void shouldHaveBody() {
			var model = callToModel();

			assertThat(model.getContent()).isEqualTo(bescheid);
		}

		private String uploadFileLink(String field) {
			return new UriTemplate("/api/binaryFiles/{vorgangId}/{field}/file").expand(VorgangHeaderTestFactory.ID, field).toString();
		}

		private String createCommandLink() {
			return new UriTemplate(CommandController.CommandByRelationController.COMMAND_BY_RELATION_PATH)
					.expand(VorgangHeaderTestFactory.ID, BescheidTestFactory.ID, BescheidTestFactory.VERSION).toString();
		}

		private String bescheidenUndSendenLink() {
			return new UriTemplate(BescheidCommandController.PATH)
					.expand(VorgangHeaderTestFactory.ID, BescheidTestFactory.ID, BescheidTestFactory.VERSION).toString();
		}

		private EntityModel<Bescheid> callToModel() {
			return callToModel(bescheid);
		}

		private EntityModel<Bescheid> callToModel(Bescheid bescheid) {
			return assembler.toModel(bescheid);
		}
	}

	@Nested
	class TestToCollectionModel {
		private final Bescheid bescheid = BescheidTestFactory.create();

		@Test
		void shouldCallToModel() {
			when(vorgangController.getVorgang(VorgangHeaderTestFactory.ID)).thenReturn(VorgangWithEingangTestFactory.create());

			callMethod();

			verify(assembler).toModel(bescheid);
		}

		@Test
		void shouldHaveEntityModel() {
			var entityModel = EntityModel.of(bescheid);
			doReturn(entityModel).when(assembler).toModel(bescheid);

			var collectionModel = callMethod();

			assertThat(collectionModel.getContent()).containsExactly(entityModel);
		}

		@Test
		void shouldHaveSelfLink() {
			when(vorgangController.getVorgang(VorgangHeaderTestFactory.ID)).thenReturn(VorgangWithEingangTestFactory.create());

			var collectionModel = callMethod();

			assertThat(collectionModel.getLinks())
					.containsExactly(linkTo(BescheidByVorgangController.class).withSelfRel());
		}

		private CollectionModel<EntityModel<Bescheid>> callMethod() {
			return assembler.toCollectionModel(List.of(bescheid));
		}

	}

	@Nested
	class TestHasVorgangCreateBescheidDocumentEnabled {

		@Nested
		class TestOnEmptyBescheidProperties {

			@BeforeEach
			void setUp() {
				when(vorgangProperties.getBescheid()).thenReturn(Collections.emptyList());
			}

			@Test
			void shouldReturnFalse() {
				var hasEnabled = callAssembler(VorgangWithEingangTestFactory.create());

				assertThat(hasEnabled).isFalse();
			}

		}

		@Nested
		class TestOnBescheidPropertiesSet {

			@BeforeEach
			void setUp() {
				when(vorgangProperties.getBescheid()).thenReturn(List.of(VorgangPropertyTestFactory.create()));
			}

			@Test
			void shouldReturnFalseIfFormEngineNameNotEquals() {
				var vorgang = createVorgang(EingangTestFactory.createBuilder()
						.header(EingangHeaderTestFactory.createBuilder()
								.formEngineName("different").build())
						.build());

				var hasEnabled = callAssembler(vorgang);

				assertThat(hasEnabled).isFalse();
			}

			@Test
			void shouldReturnTrue() {
				var hasEnabled = callAssembler(VorgangWithEingangTestFactory.create());

				assertThat(hasEnabled).isTrue();
			}
		}

		@Test
		void shouldReturnFalseOnEmptyEingang() {
			var vorgangWithEmptyEingang = createVorgang(null);

			var hasEnabled = callAssembler(vorgangWithEmptyEingang);

			assertThat(hasEnabled).isFalse();
		}

		@Test
		void shouldReturnFalseOnEmptyEingangHeader() {
			var vorgangWithEmptyEingangHeader = EingangTestFactory.createBuilder().header(null).build();

			var hasEnabled = callAssembler(createVorgang(vorgangWithEmptyEingangHeader));

			assertThat(hasEnabled).isFalse();
		}

		@Test
		void shouldReturnFalseOnEmptyFormEngineName() {
			var vorgangWithEmptyFormEngineName = createVorgang(
					EingangTestFactory.createBuilder().header(EingangHeaderTestFactory.createBuilder().formEngineName(null).build()).build());

			var hasEnabled = callAssembler(vorgangWithEmptyFormEngineName);

			assertThat(hasEnabled).isFalse();
		}

		@Test
		void shouldReturnFalseOnEmptyFormId() {
			var vorgangWithEmptyFormId = createVorgang(
					EingangTestFactory.createBuilder().header(EingangHeaderTestFactory.createBuilder().formId(null).build()).build());

			var hasEnabled = callAssembler(vorgangWithEmptyFormId);

			assertThat(hasEnabled).isFalse();
		}

		private boolean callAssembler(VorgangWithEingang vorgang) {
			return assembler.hasVorgangCreateBescheidDocumentEnabled(vorgang);
		}

		private VorgangWithEingang createVorgang(Eingang eingang) {
			return VorgangWithEingangTestFactory.createBuilder().eingang(eingang).build();
		}

	}

	@Nested
	class TestCanCreateBescheidDocumentAutomatically {

		private final VorgangWithEingang vorgangWithEingang = VorgangWithEingangTestFactory.create();

		@Test
		void shouldReturnTrueIfServiceConfigured() {
			doReturn(true).when(assembler).hasVorgangCreateBescheidDocumentEnabled(vorgangWithEingang);
			when(bescheidService.canCreateBescheidDocumentAutomatically()).thenReturn(true);

			var canCreate = assembler.canCreateBescheidDocumentAutomatically(vorgangWithEingang);

			assertThat(canCreate).isTrue();
		}

		@Test
		void shouldReturnFalseIfServiceNotConfigured() {
			when(bescheidService.canCreateBescheidDocumentAutomatically()).thenReturn(false);

			var canCreate = assembler.canCreateBescheidDocumentAutomatically(vorgangWithEingang);

			assertThat(canCreate).isFalse();
		}

		@Test
		void shouldReturnFalseIfVorgangNotConfigured() {
			doReturn(false).when(assembler).hasVorgangCreateBescheidDocumentEnabled(vorgangWithEingang);
			when(bescheidService.canCreateBescheidDocumentAutomatically()).thenReturn(true);

			var canCreate = assembler.canCreateBescheidDocumentAutomatically(vorgangWithEingang);

			assertThat(canCreate).isFalse();
		}
	}

}
