/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItem;
import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;

class DocumentHistorieProcessorTest {

	@Spy
	private DocumentHistorieProcessor processor;

	@DisplayName("Process")
	@Nested
	class TestProcess {

		private final Command command = CommandTestFactory.create();
		private final Stream<Command> commands = Stream.of(command);
		private final Command filteredCommand = CommandTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(Stream.of(filteredCommand)).when(processor).filterIrrelevantCommands(any());
		}

		@Test
		void shouldFilterIrrelevantCommands() {
			processor.process(commands);

			verify(processor).filterIrrelevantCommands(commands);
		}

		@Test
		void shouldReturnCommands() {
			var processedCommands = processor.process(commands);

			assertThat(processedCommands).containsExactly(filteredCommand);
		}
	}

	@DisplayName("Filter irrelevant commands")
	@Nested
	class TestFilterIrrelevantCommands {

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "CREATE_BESCHEID_DOCUMENT", "CREATE_BESCHEID_DOCUMENT_FROM_FILE" })
		void shouldFilterCommandWithOrder(CommandOrder order) {
			var command = CommandTestFactory.createBuilder().order(order.toString()).build();

			var commands = processor.filterIrrelevantCommands(Stream.of(command));

			assertThat(commands).isEmpty();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "CREATE_BESCHEID_DOCUMENT", "CREATE_BESCHEID_DOCUMENT_FROM_FILE" })
		void shouldKeepCommandWithOrder(CommandOrder order) {
			var command = CommandTestFactory.createBuilder().order(order.toString()).build();

			var commands = processor.filterIrrelevantCommands(Stream.of(command));

			assertThat(commands).containsExactly(command);
		}

		@DisplayName("should filter document related create attached item command")
		@Test
		void shouldFilterCreateAttachedItemCommand() {
			var command = CommandTestFactory.createBuilder().order(CommandOrder.CREATE_ATTACHED_ITEM.toString())
					.body(Map.of(VorgangAttachedItem.FIELD_ITEM_NAME, DocumentHistorieProcessor.DOCUMENT_ITEM_NAME))
					.build();

			var commands = processor.filterIrrelevantCommands(Stream.of(command));

			assertThat(commands).isEmpty();
		}
	}
}
