/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;

class PostfachNachrichtHistorieProcessorTest {

	@Spy
	private PostfachNachrichtHistorieProcessor processor;

	@Nested
	class TestProcess {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallIsNotOutgoingPostfachNachrichtByNachrichtenManager() {
			process();

			verify(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(command);
		}

		@Test
		void shouldContainCommandIfIsNotOutgoingPostfachNachrichtByNachrichtenManager() {
			doReturn(true).when(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(any());
			doReturn(command).when(processor).translateSendPostfachMailCommand(command);
			doReturn(command).when(processor).translateReceivePostfachNachrichtCommand(command);

			var commands = process();

			assertThat(commands).containsExactly(command);
		}

		@Test
		void shouldNotContainCommandIfisNotOutgoingPostfachNachrichtByNachrichtenManager() {
			doReturn(false).when(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(any());

			var commands = process();

			assertThat(commands).isEmpty();
		}

		@Test
		void shouldTranslateSendPostfachMailCommand() {
			doReturn(true).when(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(any());

			process();

			verify(processor).translateSendPostfachMailCommand(command);
		}

		@Test
		void shouldReturnTranslatedSendPostfachMailCommand() {
			var translatedCommandOrder = CommandTestFactory.create();
			doReturn(true).when(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(any());
			doReturn(translatedCommandOrder).when(processor).translateSendPostfachMailCommand(command);
			doAnswer(invocation -> invocation.getArgument(0)).when(processor).translateReceivePostfachNachrichtCommand(any());

			var commands = process();

			assertThat(commands).containsExactly(translatedCommandOrder);
		}

		@Test
		void shouldTranslateReceivePostfachNachrichtCommand() {
			doReturn(true).when(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(any());
			doAnswer(invocation -> invocation.getArgument(0)).when(processor).translateSendPostfachMailCommand(any());

			process();

			verify(processor).translateReceivePostfachNachrichtCommand(command);
		}

		@Test
		void shouldReturnTranslatedReceivePostfachNachrichtCommand() {
			var translatedCommandOrder = CommandTestFactory.create();
			doReturn(true).when(processor).isNotOutgoingPostfachNachrichtByNachrichtenManager(any());
			doAnswer(invocation -> invocation.getArgument(0)).when(processor).translateSendPostfachMailCommand(any());
			doReturn(translatedCommandOrder).when(processor).translateReceivePostfachNachrichtCommand(command);

			var commands = process();

			assertThat(commands).containsExactly(translatedCommandOrder);
		}

		private List<Command> process() {
			return processor.process(Stream.of(command)).toList();
		}
	}

	@Nested
	class TestIsNotOutgoingPostfachNachrichtenByNachrichtenManager {

		private final Command matchingCommand = CommandTestFactory.createBuilder()
				.order(CommandOrder.CREATE_ATTACHED_ITEM.name())
				.body(Map.of("item", Map.of(PostfachNachrichtHistorieProcessor.DIRECTION, PostfachNachrichtHistorieProcessor.DIRECTION_OUTGOING),
						PostfachNachrichtHistorieProcessor.CLIENT, PostfachNachrichtHistorieProcessor.OZGCLOUD_NACHRICHTEN_MANAGER))
				.build();

		@Test
		void shouldReturnFalse() {
			var result = processor.isNotOutgoingPostfachNachrichtByNachrichtenManager(matchingCommand);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnTrueOnOtherOrder() {
			var nonMatchingCommand = matchingCommand.toBuilder().order(LoremIpsum.getInstance().getWords(1)).build();

			var result = processor.isNotOutgoingPostfachNachrichtByNachrichtenManager(nonMatchingCommand);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnTrueOnIncoming() {
			var commandBody = Map.of("item",
					Map.of(PostfachNachrichtHistorieProcessor.DIRECTION, PostfachNachrichtHistorieProcessor.DIRECTION_INCOMMING),
					PostfachNachrichtHistorieProcessor.CLIENT, PostfachNachrichtHistorieProcessor.OZGCLOUD_NACHRICHTEN_MANAGER);
			var nonMatchingCommand = matchingCommand.toBuilder()
					.body(commandBody)
					.build();

			var result = processor.isNotOutgoingPostfachNachrichtByNachrichtenManager(nonMatchingCommand);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnTrueOnOtherClient() {
			var commandBody = Map.of("item",
					Map.of(PostfachNachrichtHistorieProcessor.DIRECTION, PostfachNachrichtHistorieProcessor.DIRECTION_OUTGOING),
					PostfachNachrichtHistorieProcessor.CLIENT, LoremIpsum.getInstance().getWords(1));
			var nonMatchingCommand = matchingCommand.toBuilder()
					.body(commandBody)
					.build();

			var result = processor.isNotOutgoingPostfachNachrichtByNachrichtenManager(nonMatchingCommand);

			assertThat(result).isTrue();
		}
	}

	@Nested
	class TestTranslateSendPostfachMailCommand {

		@Nested
		class TestOnHasSendPostfachMailOrder {
			@Test
			void shouldTranslateOrderToSendPostfachNachricht() {
				var command = CommandTestFactory.createBuilder().order(CommandOrder.SEND_POSTFACH_MAIL.name()).build();
				var expectedCommand = command.toBuilder().order(CommandOrder.SEND_POSTFACH_NACHRICHT.name()).build();

				var resultCommand = processor.translateSendPostfachMailCommand(command);

				assertThat(resultCommand).usingRecursiveComparison().isEqualTo(expectedCommand);
			}
		}

		@Nested
		class TestOnHasNotSendPostfachMailOrder {
			@Test
			void shouldReturnSameCommand() {
				var command = CommandTestFactory.create();

				var resultCommand = processor.translateSendPostfachMailCommand(command);

				assertThat(resultCommand).isSameAs(command);
			}
		}
	}

	@Nested
	class TestTranslateReceivePostfachNachrichtCommand {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallIsIncomingNachrichtbyNachrichtenManager() {
			processor.translateReceivePostfachNachrichtCommand(command);

			verify(processor).isIncomingNachrichtByNachrichtenManager(command);
		}

		@Test
		void shouldTranslateOnIsIncomingNachrichtbyNachrichtenManager() {
			doReturn(true).when(processor).isIncomingNachrichtByNachrichtenManager(any());
			var expectedCommand = command.toBuilder().order(CommandOrder.RECEIVE_POSTFACH_NACHRICHT.name()).build();

			var resultCommand = processor.translateReceivePostfachNachrichtCommand(command);

			assertThat(resultCommand).usingRecursiveComparison().isEqualTo(expectedCommand);
		}

		@Test
		void shouldNotTranslateOnIsNotIncomingNachrichtbyNachrichtenManager() {
			doReturn(false).when(processor).isIncomingNachrichtByNachrichtenManager(any());

			var resultCommand = processor.translateReceivePostfachNachrichtCommand(command);

			assertThat(resultCommand).isSameAs(command);
		}
	}

	@Nested
	class TestIsIncomingNachrichtByNachrichtenManager {

		private final Command matchingCommand = CommandTestFactory.createBuilder()
				.order(CommandOrder.CREATE_ATTACHED_ITEM.name())
				.body(Map.of("item", Map.of(PostfachNachrichtHistorieProcessor.DIRECTION, PostfachNachrichtHistorieProcessor.DIRECTION_INCOMMING),
						PostfachNachrichtHistorieProcessor.CLIENT, PostfachNachrichtHistorieProcessor.OZGCLOUD_NACHRICHTEN_MANAGER))
				.build();

		@Test
		void shouldReturnTrue() {
			var result = processor.isIncomingNachrichtByNachrichtenManager(matchingCommand);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseOnOtherOrder() {
			var nonMatchingCommand = matchingCommand.toBuilder().order(LoremIpsum.getInstance().getWords(1)).build();

			var result = processor.isIncomingNachrichtByNachrichtenManager(nonMatchingCommand);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnFalseOnOutgoing() {
			var commandBody = Map.of("item",
					Map.of(PostfachNachrichtHistorieProcessor.DIRECTION, PostfachNachrichtHistorieProcessor.DIRECTION_OUTGOING),
					PostfachNachrichtHistorieProcessor.CLIENT, PostfachNachrichtHistorieProcessor.OZGCLOUD_NACHRICHTEN_MANAGER);
			var nonMatchingCommand = matchingCommand.toBuilder()
					.body(commandBody)
					.build();

			var result = processor.isIncomingNachrichtByNachrichtenManager(nonMatchingCommand);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnFalseOnOtherClient() {
			var commandBody = Map.of("item",
					Map.of(PostfachNachrichtHistorieProcessor.DIRECTION, PostfachNachrichtHistorieProcessor.DIRECTION_INCOMMING),
					PostfachNachrichtHistorieProcessor.CLIENT, LoremIpsum.getInstance().getWords(1));
			var nonMatchingCommand = matchingCommand.toBuilder()
					.body(commandBody)
					.build();

			var result = processor.isIncomingNachrichtByNachrichtenManager(nonMatchingCommand);

			assertThat(result).isFalse();
		}
	}
}
