/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang;

import static de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.time.format.DateTimeFormatter;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.clientattribute.ClientAttribute;
import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserIdMapper;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageTestFactory;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttribute;
import de.ozgcloud.vorgang.grpc.clientAttribute.GrpcClientAttributeValue;

class VorgangHeaderMapperTest {

	@Spy
	@InjectMocks
	private VorgangHeaderMapper mapper = Mappers.getMapper(VorgangHeaderMapper.class);

	@Spy
	private UserIdMapper userIdMapper = Mappers.getMapper(UserIdMapper.class);

	@Nested
	class TestToVorgangHeader {

		@Test
		void shouldMapCreatedAt() {
			var header = callMapper();

			assertThat(header.getCreatedAt()).isEqualTo(CREATED_AT);
		}

		@Test
		void shouldMapNummer() {
			var header = callMapper();

			assertThat(header.getNummer()).isEqualTo(NUMMER);
		}

		@Test
		void shouldMapAktenzeichen() {
			var header = callMapper();

			assertThat(header.getAktenzeichen()).isEqualTo(AKTENZEICHEN);
		}

		@Test
		void shouldMapStatus() {
			var header = callMapper();

			assertThat(header.getStatus()).isEqualTo(VorgangStatus.NEU);
		}

		@Test
		void shouldMapAssignedTo() {
			var header = callMapper();

			assertThat(header.getAssignedTo()).isEqualTo(UserId.from(ASSIGNED_TO));
		}

		@Nested
		class NextFrist {

			private final GrpcClientAttributeValue value = GrpcClientAttributeValue.newBuilder()
					.setStringValue(WiedervorlageTestFactory.FRIST.format(DateTimeFormatter.ISO_DATE)).build();

			@Test
			void shouldSetEmptyOptionalIfNotExist() {
				var vorgangHeaderWithEmptyClientAttributes = GrpcVorgangHeaderTestFactory.createBuilder().clearClientAttributes().build();

				var header = mapper.toVorgangHeader(vorgangHeaderWithEmptyClientAttributes);

				assertThat(header.getNextFrist()).isNull();
			}

			@Test
			void shouldNotSetIfWrongClient() {
				var clientAttribute = GrpcClientAttribute.newBuilder()
						.setAttributeName(ClientAttribute.WIEDERVORLAGE_NEXT_FRIST.getAttributeName())
						.setValue(value)
						.build();
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder()
						.addClientAttributes(clientAttribute).build();

				var header = mapper.toVorgangHeader(vorgangHeader);

				assertThat(header.getNextFrist()).isNull();

			}

			@Test
			void shouldSetIfExists() {
				var clientAttribute = GrpcClientAttribute.newBuilder()
						.setAttributeName(ClientAttribute.WIEDERVORLAGE_NEXT_FRIST.getAttributeName())
						.setClientName(ClientAttribute.WIEDERVORLAGE_NEXT_FRIST.getClientName().toString())
						.setValue(value)
						.build();
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder()
						.addClientAttributes(clientAttribute).build();

				var header = mapper.toVorgangHeader(vorgangHeader);

				assertThat(header.getNextFrist()).isEqualTo(WiedervorlageTestFactory.FRIST);
			}

		}

		@Nested
		class HasPostfachNachricht {
			@Test
			void shouldSetFalseIfNotExists() {
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().clearClientAttributes().build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.isHasPostfachNachricht()).isFalse();
			}

			@Test
			void shouldSetFalseForWrongClient() {
				var clientAttribute = GrpcClientAttributeTestFactory.createBuilder()
						.setValue(GrpcClientAttributeValue.newBuilder().setBoolValue(true).build())
						.build();
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().addClientAttributes(clientAttribute).build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.isHasPostfachNachricht()).isFalse();
			}

			@ParameterizedTest
			@ValueSource(booleans = { true, false })
			void shouldBeValueOfClientAttribute(boolean value) {
				var clientAttribute = GrpcClientAttributeTestFactory.createWith(ClientAttribute.HAS_POSTFACH_NACHRICHT, value);
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().addClientAttributes(clientAttribute).build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.isHasPostfachNachricht()).isEqualTo(value);
			}
		}

		@Nested
		class HasNewPostfachNachricht {
			@Test
			void shouldSetFalseIfNotExists() {
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().clearClientAttributes().build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.isHasNewPostfachNachricht()).isFalse();
			}

			@Test
			void shouldSetFalseForWrongClient() {
				var clientAttribute = GrpcClientAttributeTestFactory.createBuilder()
						.setValue(GrpcClientAttributeValue.newBuilder().setBoolValue(true).build())
						.build();
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().addClientAttributes(clientAttribute).build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.isHasPostfachNachricht()).isFalse();
			}

			@ParameterizedTest
			@ValueSource(booleans = { true, false })
			void shouldBeValueOfClientAttribute(boolean value) {
				var clientAttribute = GrpcClientAttributeTestFactory.createWith(ClientAttribute.HAS_NEW_POSTFACH_NACHRICHT, value);
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().addClientAttributes(clientAttribute).build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.isHasNewPostfachNachricht()).isEqualTo(value);
			}
		}

		@Nested
		class TestAntragBewilligt {
			@Test
			void shouldSetNull() {
				var vorgangHeaderWithEmptyClientAttributes = GrpcVorgangHeaderTestFactory.createBuilder().clearClientAttributes().build();

				var header = mapper.toVorgangHeader(vorgangHeaderWithEmptyClientAttributes);

				assertThat(header.getAntragBewilligt()).isNull();
			}

			@ParameterizedTest
			@ValueSource(booleans = { true, false })
			void shouldBeValueOfClientAttribute(boolean value) {
				var clientAttribute = GrpcClientAttributeTestFactory.createWith(ClientAttribute.ANTRAG_BEWILLIGT, value);
				var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().addClientAttributes(clientAttribute).build();

				var mapped = mapper.toVorgangHeader(vorgangHeader);

				assertThat(mapped.getAntragBewilligt()).isEqualTo(value);
			}
		}

		private VorgangHeader callMapper() {
			return mapper.toVorgangHeader(GrpcVorgangHeaderTestFactory.create());
		}
	}

	@Nested
	class TestArchiving {

		@Test
		void shouldBeFalseIfClientAttributeIsMissing() {
			var header = mapper.toVorgangHeader(GrpcVorgangHeaderTestFactory.create());

			assertThat(header.isArchiving()).isFalse();
		}

		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldBeValueOfClientAttribute(boolean value) {
			var clientAttribute = GrpcClientAttributeTestFactory.createWith(ClientAttribute.ARCHIVING, value);
			var vorgangHeader = GrpcVorgangHeaderTestFactory.createBuilder().addClientAttributes(clientAttribute).build();

			var mapped = mapper.toVorgangHeader(vorgangHeader);

			assertThat(mapped.isArchiving()).isEqualTo(value);
		}
	}
}
