/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;

import de.ozgcloud.alfa.common.LinkedResourceProcessor;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.statistic.Statistic;
import de.ozgcloud.alfa.statistic.StatisticTestFactory;
import lombok.SneakyThrows;

class VorgangModelAssemblerTest {

	@Spy
	@InjectMocks
	private VorgangModelAssembler modelAssembler;
	@Mock
	private CurrentUserService userService;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private LinkedResourceProcessor<Vorgang> linkedResourceProcessor;

	private static final int PAGE_SIZE = 100;

	@DisplayName("To collectionModel")
	@Nested
	class TestCollectionModel {

		@DisplayName("init model")
		@Nested
		class TestInitModel {

			@Test
			void shouldBeCalled() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().searchBy(Optional.empty()).build();
				var response = VorgaengeHeaderResponse.builder().vorgaengeHeader(Collections.emptyList()).total(PAGE_SIZE).build();
				var entities = Stream.of(VorgangHeaderTestFactory.create());
				var statistic = StatisticTestFactory.create();

				modelAssembler.toCollectionModel(entities, response, requestCriteria, statistic);

				verify(modelAssembler).initModel(entities, statistic);
			}
		}

		@DisplayName("next link")
		@Nested
		class TestNextLink {

			@Test
			void shouldAddIfPagesLeft() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().searchBy(Optional.empty()).build();
				var response = VorgaengeHeaderResponse.builder().vorgaengeHeader(Collections.emptyList()).total(PAGE_SIZE).build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), response, requestCriteria,
						StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_NEXT).get()).extracting(Link::getHref)
						.isEqualTo(VorgangController.PATH + "?page=1&limit=100&assignedTo=" + UserProfileTestFactory.ID.toString() + "&status=NEU");
			}

			@Test
			void shouldNotAddIfLastPage() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.create();

				var response = VorgaengeHeaderResponse.builder()
						.vorgaengeHeader(List.of(VorgangHeaderTestFactory.create()))
						.total(1).build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), response, requestCriteria,
						StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_NEXT)).isNotPresent();
			}
		}

		@DisplayName("previous link")
		@Nested
		class TestPreviousLink {

			@Test
			void shouldAddIfNotFirstPage() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder()
						.page(1)
						.limit(1)
						.searchBy(Optional.empty())
						.build();
				var response = VorgaengeHeaderResponse.builder()
						.vorgaengeHeader(List.of(VorgangHeaderTestFactory.create()))
						.total(200)
						.build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create(), VorgangHeaderTestFactory.create()),
						response,
						requestCriteria, StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_PREVIOUS).get()).extracting(Link::getHref)
						.isEqualTo(VorgangController.PATH + "?page=0&limit=1&assignedTo=" + UserProfileTestFactory.ID.toString() + "&status=NEU");
			}

			@Test
			void shouldNotAddIfFirstPage() {
				var requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().page(0).build();
				var response = VorgaengeHeaderResponse.builder().vorgaengeHeader(Collections.emptyList()).total(PAGE_SIZE).build();

				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create()), response, requestCriteria,
						StatisticTestFactory.create());

				assertThat(model.getLink(VorgangModelAssembler.REL_PREVIOUS)).isNotPresent();
			}
		}

		@DisplayName("add self link")
		@Nested
		class TestSelfLink {

			final FindVorgaengeHeaderRequestCriteria requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder()
					.page(1)
					.limit(1)
					.build();
			final VorgaengeHeaderResponse response = VorgaengeHeaderResponse.builder()
					.vorgaengeHeader(List.of(VorgangHeaderTestFactory.create()))
					.total(200)
					.build();

			@Test
			void shouldAddLink() {
				var model = modelAssembler.toCollectionModel(Stream.of(VorgangHeaderTestFactory.create(), VorgangHeaderTestFactory.create()),
						response,
						requestCriteria, StatisticTestFactory.create());

				assertThat(model.getLink(IanaLinkRelations.SELF_VALUE)).isPresent();
				assertThat(model.getLink(IanaLinkRelations.SELF_VALUE).get()).extracting(Link::getHref).isEqualTo(VorgangController.PATH);
			}
		}
	}

	@Nested
	class TestInitModel {

		private static final Link DUMMY_LINK = Link.of("dummy");
		private final VorgangHeader vorgangHeader = VorgangHeaderTestFactory.create();
		private final Stream<VorgangHeader> entities = Stream.of(vorgangHeader);
		private final Statistic statistic = StatisticTestFactory.create();
		private final EntityModel<Vorgang> entityModel = EntityModel.of(vorgangHeader);
		private final EntityModel<Vorgang> processedEntityModel = EntityModel.of(vorgangHeader, DUMMY_LINK);

		@BeforeEach
		void mock() {
			doReturn(entityModel).when(modelAssembler).toModel(any());
			when(linkedResourceProcessor.process(any())).thenReturn(processedEntityModel);
		}

		@Test
		void shouldHaveSelfLink() {
			var model = modelAssembler.initModel(entities, statistic);

			assertThat(model.getLink(IanaLinkRelations.SELF_VALUE)).get().extracting(Link::getHref).isEqualTo(VorgangController.PATH);
		}

		@Test
		void shouldCallToModel() {
			modelAssembler.initModel(entities, statistic);

			verify(modelAssembler).toModel(vorgangHeader);
		}

		@Test
		void shouldCallLinkedResourceProcessor() {
			modelAssembler.initModel(entities, statistic);

			verify(linkedResourceProcessor).process(entityModel);
		}

		@Test
		@SneakyThrows
		void shouldContainProcessedEntity() {
			var mapper = new ObjectMapper().registerModules(new JavaTimeModule(), new Jdk8Module());

			var model = mapper.writeValueAsString(modelAssembler.initModel(entities, statistic));

			assertThat(model).contains(mapper.writeValueAsString(processedEntityModel));
		}
	}

	@DisplayName("Build page link")
	@Nested
	class TestBuildPageLink {

		private final String linkRel = "dummyLinkRel";
		final FindVorgaengeHeaderRequestCriteria requestCriteria = FindVorgaengeRequestCriteriaTestFactory.createBuilder().build();

		@Test
		void shouldContainsPage() {
			var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

			assertThat(link.getHref()).contains(VorgangController.PARAM_PAGE + "=0");
		}

		@Test
		void shouldContainsLimit() {
			var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

			assertThat(link.getHref()).contains(VorgangController.PARAM_LIMIT + "=" + FindVorgaengeRequestCriteriaTestFactory.LIMIT);
		}

		@Test
		void shouldHaveLinkRel() {
			var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

			assertThat(link.hasRel(linkRel)).isTrue();
		}

		@DisplayName("search by")
		@Nested
		class TestSearchBy {

			@Test
			void shouldAddAsParamIfExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

				assertThat(link.getHref())
						.contains(VorgangController.PARAM_SEARCH + "="
								+ FindVorgaengeRequestCriteriaTestFactory.SEARCH_BY.replace(StringUtils.SPACE, "%20"));
			}

			@Test
			void shouldNOTaddIfNotExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().searchBy(Optional.empty()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_SEARCH);
			}
		}

		@DisplayName("assignedTo")
		@Nested
		class TestAssignedTo {

			@Test
			void shouldAddAsParamIfShouldFilter() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

				assertThat(link.getHref()).contains(VorgangController.PARAM_ASSIGNED_TO + "=" + UserProfileTestFactory.ID);
			}

			@Test
			void shouldNOTaddIfShouldNotFilter() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().filterCriteria(
						FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false).build()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_ASSIGNED_TO);
			}

			@DisplayName("getAssignedToValue")
			@Nested
			class TestGetAssignedToValue {

				@Test
				void shouldReturnValueIfExists() {
					var assignedToValue = modelAssembler.getAssignedToValue(FilterCriteriaTestFactory.create());

					assertThat(assignedToValue).isEqualTo(UserProfileTestFactory.ID.toString());
				}

				@Test
				void shouldReturnEmptyStringIfNotExists() {
					var assignedToValue = modelAssembler
							.getAssignedToValue(FilterCriteriaTestFactory.createBuilder().assignedTo(Optional.empty()).build());

					assertThat(assignedToValue).isEqualTo(StringUtils.EMPTY);
				}
			}
		}

		@DisplayName("status")
		@Nested
		class TestStatus {

			@Test
			void shouldAddAsParamIfExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria);

				assertThat(link.getHref()).contains(VorgangController.PARAM_STATUS + "=" + FilterCriteriaTestFactory.VORGANG_STATUS);
			}

			@Test
			void shouldNOTaddIfNotExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().filterCriteria(
						FilterCriteriaTestFactory.createBuilder().status(Optional.empty()).build()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_STATUS);
			}
		}

		@DisplayName("next wiedervorlage frist")
		@Nested
		class TestWiedervorlagen {

			@Test
			void shouldAddAsParamExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder()
						.filterCriteria(FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(true).build()).build());

				assertThat(link.getHref())
						.contains(VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST + "=" + VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST_EXISTS);
			}

			@Test
			void shouldNOTaddIfNotExists() {
				var link = modelAssembler.buildPageLink(0, linkRel, requestCriteria.toBuilder().filterCriteria(
						FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(false).build()).build());

				assertThat(link.getHref()).doesNotContain(VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST);
			}
		}
	}

	@Nested
	class TestLinksOnModel {

		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@Test
		void shouldHaveSelfLink() {
			var link = toModel().getLink(IanaLinkRelations.SELF_VALUE);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(VorgangController.PATH + "/" + VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldHaveVorgangMitEingangLink() {
			var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_MIT_EINGANG);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(VorgangController.PATH + "/" + VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldHaveAssignLink() {
			doReturn(true).when(modelAssembler).isAssigneable(vorgang);

			var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(buildCommandLink());
		}

		@Test
		void shouldNotHaveAssignLink() {
			doReturn(false).when(modelAssembler).isAssigneable(vorgang);

			var link = toModel().getLink(VorgangModelAssembler.REL_VORGANG_ASSIGN);

			assertThat(link).isEmpty();
		}

		@Test
		void shouldHaveWiedervorlagenLink() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);

			var link = toModel().getLink(VorgangModelAssembler.REL_WIEDERVORLAGEN);

			assertThat(link).isPresent().get().extracting(Link::getHref)
					.isEqualTo("/api/wiedervorlages?vorgangId=" + VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldNotHaveWiedervorlagenLink() {
			var link = toModel().getLink(VorgangModelAssembler.REL_WIEDERVORLAGEN);

			assertThat(link).isEmpty();
		}

		private String buildCommandLink() {
			return String.format("/api/vorgangs/%s/relations/%s/%s/commands", VorgangHeaderTestFactory.ID,
					VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.VERSION);
		}

		private EntityModel<Vorgang> toModel() {
			return modelAssembler.toModel(vorgang);
		}
	}

	@Nested
	class TestIsAssigneable {

		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@Test
		void shoudCheckIfVorgangIsEditable() {
			givenVerwaltungUser();

			modelAssembler.isAssigneable(vorgang);

			verify(vorgangService).isVorgangEditable(vorgang);
		}

		@Test
		void shouldReturnTrueForVerwaltungUser() {
			givenVerwaltungUser();
			givenVorgangIsEditable();

			assertThat(modelAssembler.isAssigneable(vorgang)).isTrue();
		}

		@Test
		void shouldReturnTrueForPoststelleUser() {
			givenNotVerwaltungUser();
			givenPoststelleUser();
			givenVorgangIsEditable();

			assertThat(modelAssembler.isAssigneable(vorgang)).isTrue();
		}

		@Test
		void shouldReturnFalseForNotPoststelleUser() {
			givenNotVerwaltungUser();
			givenNotPoststelleUser();

			assertThat(modelAssembler.isAssigneable(vorgang)).isFalse();
		}

		@Test
		void shouldReturnFalseForVerwaltungUserAndVorgangNotEditable() {
			givenVerwaltungUser();
			givenVorgangIsNotEditable();

			var assignable = modelAssembler.isAssigneable(vorgang);

			assertThat(assignable).isFalse();
		}

		@Test
		void shouldReturnFalseForPoststelleUserAndVorgangNotEditable() {
			givenNotVerwaltungUser();
			givenPoststelleUser();
			givenVorgangIsNotEditable();

			var assignable = modelAssembler.isAssigneable(vorgang);

			assertThat(assignable).isFalse();
		}

		private void givenVerwaltungUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);
		}

		private void givenNotVerwaltungUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(false);
		}

		private void givenPoststelleUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(true);
		}

		private void givenNotPoststelleUser() {
			when(userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(false);
		}

		private void givenVorgangIsEditable() {
			doReturn(true).when(vorgangService).isVorgangEditable(vorgang);
		}

		private void givenVorgangIsNotEditable() {
			doReturn(false).when(vorgangService).isVorgangEditable(vorgang);
		}
	}
}